# Copyright (C) 2008 Ansgar Wiechers <ansgar.wiechers@planetcobalt.net>
#
# This module is free software. You may use, redistribute and/or modify it
# under the terms of the GNU General Public License Version 2.0 as found at
# <http://www.gnu.org/licenses/old-licenses/gpl-2.0.html>.

package BerkeleyDB;
use FilterDB::DB;
our @ISA = qw( DB );

=head1 NAME

  FilterDB::BerkeleyDB -- Backscatter filter backend (BerkeleyDB)

=head1 SYNOPSIS

  use FilterDB::BerkeleyDB;
  use strict;

  my $db = new BerkeleyDB(
      'filename' => '/PATH/TO/my.db',
    );  

  if ($db->check('foo')) {
    print "'foo' found\n";
  } else {
    print "'foo' not found\n";
  }

  $db->insert('foo');
  $db->insert('bar');

  if ($db->check('foo')) {
    print "'foo' found\n";
  } else {
    print "'foo' not found\n";
  }

  my %records = $db->list();
  my @keys = keys %records;
  if ($#keys >= 0) {
    print "$records{$key}: $key\n" foreach my $key (@keys);
  }

=head1 REQUIRES

Perl5.8.8, BerkeleyDB, Date::Manip

=head1 DESCRIPTION

The BerkeleyDB class implements a BerkeleyDB backend for the backscatter filter.
It stores MD5 hashes of given values and the timestamps when the values were
entered into the database. It provides methods to check whether the hash for any
given value is present in the database, to add insert new values, to list the
database contents, and to remove outdated values.

Status information will be logged to syslog if the calling application has
opened the syslog via openlog().

=head1 METHODS

=head2 Creation

=over 4

=item new($filename, [$max_age])

Creates a BerkeleyDB instance from the given $filename. If the file isn't
present it will be created automatically. The optional argument $max_age
specifies the number of days after which an entry should be removed from the
database (by the cleanup() function). The default is 4 days.

=back

=head2 Access

=over 4

=item check($value)

Checks whether the MD5 hash of the given $value is present in the database.
Returns TRUE if the hash is present, otherwise FALSE.

=item cleanup()

Removes entries older than $max_age days from the database.

=item insert($value)

Inserts the MD5 hash of the given $value and the current timestamp into the
database. If the hash is already present, the existing entry is updated with the
durrent timestamp.

=item list()

Returns a dictionary with all MD5 hash->timestamp pairs stored in the database.

=back

=head1 AUTHOR

Ansgar Wiechers <ansgar.wiechers@planetcobalt.net>

=cut

require 5.8.8;

use strict;
use warnings;
use BerkeleyDB;
use Date::Manip;
use Digest::MD5 qw( md5_hex );
use Sys::Syslog qw( :standard :macros );

use constant TRUE  => 1;
use constant FALSE => not TRUE;

my $env_flags = DB_INIT_MPOOL | DB_INIT_LOCK | DB_INIT_LOG | DB_INIT_TXN | DB_CREATE;
my $db_flags = DB_CREATE;
my $default_max_age = 4; # days

sub new {
	my ($class, %args) = @_;
	my $self = bless { %args }, $class;

	_terminate('Error creating BerkeleyDB instance: No database file specified') unless defined $self->{filename};
	$self->{max_age} = $default_max_age unless defined $self->{max_age} and $self->{max_age} =~ m/^[0-9]+$/;

	syslog(LOG_DEBUG, "Creating BerkeleyDB instance [filename=>$self->{filename}, max_age=>$self->{max_age}]");

	$self->{env} = new BerkeleyDB::Env
			-Home    => $ENV{'HOME'},
			-Flags   => $env_flags,
			-ErrFile => "$ENV{'HOME'}/db_error.log"
		or _terminate("Cannot create database environment: $! $BerkeleyDB::Error\n");

	my $txn = $self->{env}->txn_begin();
	$self->{db} = new BerkeleyDB::Hash
			-Env      => $self->{env},
			-Filename => $self->{filename},
			-Flags    => $db_flags,
			-Txn      => $txn
		or _terminate("Cannot open file $self->{filename}: $! $BerkeleyDB::Error\n");
	$txn->txn_commit();

	syslog(LOG_DEBUG, "BerkeleyDB instance created");

	return $self;
}

sub DESTROY {
	my ($self) = @_;
	syslog(LOG_DEBUG, "Destroying BerkeleyDB instance");
	undef $self->{db};
}

sub check {
	my ($self, $value) = @_;
	my $found = FALSE;
	my $created;

	_terminate('check(): Missing argument $value') unless defined $value;

	syslog(LOG_DEBUG, "Checking for presence of '$value'");

	my $txn = $self->{env}->txn_begin();
	my $cursor = $self->{db}->db_cursor();
	$found = TRUE if $cursor->c_get(md5_hex($value), $created, DB_SET) == 0;
	$cursor->c_close();
	undef $cursor ;
	$txn->txn_commit();

	return $found;
}

sub cleanup {
	my ($self) = @_;
	my ($hash, $created) = ('', '');
	my $limit = DateCalc("today", "-$self->{max_age} days");

	syslog(LOG_DEBUG, "Removing entries older than $self->{max_age} days");

	my $txn = $self->{env}->txn_begin();
	my $cursor = $self->{db}->db_cursor();
	while ($cursor->c_get($hash, $created, DB_NEXT) == 0) {
		$cursor->c_del() if Date_Cmp($created, $limit) < 0;
	}
	$cursor->c_close();
	undef $cursor;
	$txn->txn_commit();
}

sub insert {
	my ($self, $value) = @_;

	_terminate('insert(): Missing argument $value') unless defined $value;

	syslog(LOG_DEBUG, "Inserting '$value' into database");

	my $txn = $self->{env}->txn_begin();
	$self->{db}->db_put(md5_hex($value), ParseDate('today'));
	$txn->txn_commit();
}

sub list {
	my ($self) = @_;
	my ($hash, $created) = ('', '');
	my %records;

	syslog(LOG_DEBUG, 'Listing database entries');

	my $txn = $self->{env}->txn_begin();
	my $cursor = $self->{db}->db_cursor();
	while ($cursor->c_get($hash, $created, DB_NEXT) == 0) {
		$records{$hash} = $created;
	}
	$cursor->c_close();
	undef $cursor ;
	$txn->txn_commit();

	return %records;
}

sub _terminate($) {
	syslog(LOG_ERR, $_[0]);
	die $_[0];
}

1;
